# -*- coding: utf-8 -*-
'''
pypos3d intends to be a highlevel library for:
  - WaveFront Files (3D geometry files .OBJ)
  - Poser Files     (3D Scene description files .PZ3 .PZZ .CR2 .CRZ)

pypos3d aims to replace Pojamas.DATA3D library.
Application ePOSER (in Eclipse RCP) is replaced by a new MMI concept
based on LibreOffice/Python.

For 3D vizualisation: install and use pypos3dv
This package is distributed appart from pypos3d, because it needs some 'heavy' 
dependencies : PyOpenG, PyGLM and GLFW.
I don't want 'pypos3d' to depend on them.

'''
__version__ = '1.3.1'
__author__  = 'Olivier DUFAILLY'
__license__ = 'BSD'
