# -*- coding: utf-8 -*-
'''
  pypos3d.wftk intends to be a highlevel library for:
  - WaveFront Files (3D geometry files)

  Requires : numpy, scipy.spatial
'''
__version__ = '1.5.0'
__author__  = 'Olivier DUFAILLY'
__license__ = 'BSD'
